from collections import defaultdict
from itertools import chain, permutations

from symbols import Symbol, Symbols, escape


class Rule(Symbols):
    def __init__(self, rule=None, left=None, right=None):
        if rule is None:
            self.left = Symbol(left)
        else:
            if isinstance(rule, Rule):
                self.left = Symbol(rule.left)
                if right is None:
                    right = rule
            else:
                try:
                    left_side, right_side = rule.split(' → ')
                except ValueError:
                    raise ValueError('Multiple rules stuck together? ' + rule)
                self.left = Symbol(left_side)
                if right is None:
                    right = right_side
            if left:
                self.left.update(Symbol(left))
        Symbols.__init__(self, right)

        self.epsilon_rules = list(self._process_optional_symbols())
        self.attr_matches = self._process_matches()

    def __add__(self, value):
        return Rule(str(self) + str(value))

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash(str(self))

    def __str__(self):
        return str(self.left) + ' → ' + Symbols.__str__(self)

    def __repr__(self):
        return 'Rule(' + escape(self) + ')'

    # TODO: byla to dobrá lekce, ale hodilo by se spíš se vrátit k __nonzero__,
    # které Python definoval pro Rule jako "is not None"; teď mi to rozděluje
    # pravidla od ε-pravidel, což vůbec nepotřebuju

    def _process_matches(self):
        attr_matches = defaultdict(list)  # → OrderedDict, abych nemusel řadit
        for index, token in enumerate([self.left] + list(self), -1):
            for attr, value in token.items():
                if value is None:
                    attr_matches[attr].append(index)

        for attr, match_on_indexes in attr_matches.items():
            if len(match_on_indexes) == 1:
                raise ValueError('No other symbol to match "%s" on in %s'
                                 % (attr, self))
        return attr_matches

    def _process_optional_symbols(self):
        for index, token in enumerate(self, 1):
            if token.get('optional', False):
                if 'phrase' not in token:
                    new_token = Symbol(token)
                    del new_token['optional']

                    # token['word'] = None
                    token.pop('word', None)
                    token.pop('lemma', None)
                    token.pop('tag', None)
                    token['phrase'] = self.left['phrase'] + str(index)

                    new_head = Symbol(token)
                    del new_head['optional']
                    yield Rule(left=new_head, right=[new_token])
                yield Rule(left=Symbol(phrase=token['phrase']), right=[])

    def insert_symbol(self, inserted_symbol, before=False, inside=True,
                      after=False, position=None):
        if isinstance(inserted_symbol, str):
            inserted_symbol = Symbol(inserted_symbol)

        if position is not None:
            new_right = list(self)
            new_right.insert(position, inserted_symbol)
        else:
            new_right = [inserted_symbol] if before else []
            if inside:
                for symbol in self:
                    new_right.extend((symbol, inserted_symbol))
                if not after:
                    new_right.pop()
            else:
                new_right.extend(self)
                if after:
                    new_right.append(inserted_symbol)
        return Rule(left=self.left, right=new_right)

    def permute(self):
        for permutation in permutations(self):
            yield Rule(left=self.left, right=permutation)


def permute_rules(rules):
    return chain.from_iterable(Rule(rule).permute() for rule in rules)


def insert_into_rules(rules, inserted_symbol, before=False, inside=True,
                      after=False, position=None):
    for rule in rules:
        yield Rule(rule).insert_symbol(
            inserted_symbol, before, inside, after, position)
