from itertools import chain
import re

from log import log
from shlex import quote, shlex  # shlex je teď můj


ORDER = 'kegncpamdxytzw~'
FIXED_ATTRS = ('word', 'lemma', 'tag')


class Symbol(dict):
    def __init__(self, iterable=None, *args, **kwargs):
        if isinstance(iterable, str):
            symbols = read_rule(iterable)
            try:
                attrs = next(symbols)
            except StopIteration:
                raise ValueError('No symbol passed: "' + iterable + '"')
            # '•' neprojde
            self.update(attrs)
            extra = list(symbols)
            if extra:
                # TODO: taky ValueError?
                log.warning('Extra symbols passed, ignoring them: %s',
                            ' '.join(str(t) for t in extra))
        elif iterable is not None:
            super().__init__(iterable, *args, **kwargs)
        else:
            super().__init__(*args, **kwargs)

    def __str__(self, show_tag=False):
        if 'structural' in self:
            return self.stringify_structural()
        attrs = dict(self)
        attrs.pop('line', None)
        phrase_head = attrs.pop('phrase', None)
        phrase_head = [(phrase_head, None)] if phrase_head else []
        optional = '?' if attrs.pop('optional', None) else ''

        if not show_tag:
            _ = attrs.pop('tag', None)
        fixed_attrs = [(attr, attrs[attr]) for attr in FIXED_ATTRS
                       if attr in attrs]
        short_attrs = [(attr, attrs[attr]) for attr in ORDER if attr in attrs]
        other_attrs = sorted((attr, val) for (attr, val) in attrs.items()
                             if attr not in FIXED_ATTRS and (
                                len(attr) > 1 or attr not in ORDER))
        attrs = (attr + '=' + escape(val) if val is not None else attr
                 for attr, val in chain(phrase_head, fixed_attrs, short_attrs,
                                        other_attrs))
        return '({symbol})'.format(symbol=' '.join(attrs)) + optional

    def __repr__(self):
        return 'Symbol(' + escape(self) + ')'

    def match(self, chart_edge):
        for attr, indexes in sorted(chart_edge.rule.attr_matches.items()):
            value = None
            for index in indexes:
                if index == -1:
                    continue
                if index < chart_edge.position:
                    if attr not in chart_edge[index]:
                        log.warning('             %s not in %s', attr,
                                    chart_edge[index])
                        continue
                    value = chart_edge[index][attr]
                elif value is not None and index == chart_edge.position:
                    if attr not in self:
                        log.warning('not present  %s %s', attr,
                                    str(self))
                    elif value is None:
                        log.debug('attr null    %s in %s', attr, self)
                    # TODO: přesunout funkci? (už se to vypisuje)
                    elif value != self[attr]:
                        log.error('mismatch     %s %s != %s in %s and %s',
                                  attr, self[attr], value, self, chart_edge)
                        return
        return True

    def stringify_structural(self):
        closing = '/' if self.get('closing') else ''
        tag = self.get('structural')
        empty = '/' if self.get('empty', '') else ''
        attrs = ''.join(
            ' {}="{}"'.format(attr, val) for attr, val in sorted(self.items())
             if attr not in ('line', 'closing', 'structural', 'empty'))
        return '<{closing}{tag}{attrs}{empty}>'.format(
            closing=closing, tag=tag, attrs=attrs, empty=empty)

    def vertical(self):
        return '{}\t{}\t{}'.format(self['word'], self['lemma'], self['tag'])


class Symbols(list):
    """Abstrakce pro pravou stranu pravidla/hrany (anebo její část)"""
    def __init__(self, iterable=None, position=None):
        self.position = position
        if isinstance(iterable, str):
            self.extend(read_rule(iterable))
        elif iterable is not None:
            super().__init__((Symbol(s) for s in iterable))
        else:
            super().__init__()
        if self and position is not None:
            self.parse_position()

    def __add__(self, value):
        if isinstance(value, Symbol):
            self.append(value)
        elif isinstance(value, str):
            self.extend(Symbols(value))
        else:
            return super().__add__(value)

    def __str__(self):
        if not self:
            return 'ε' if self.position is None else 'ε •'
        symbols = [str(symbol) for symbol in self]
        if self.position is not None:
            symbols.insert(self.position, '•')
        return ' '.join(symbols)

    def __repr__(self):
        return 'Symbols(' + escape(self) + ')'

    def parse_position(self):
        for index, symbol in enumerate(list(self)):
            if symbol == '•':
                if self.position is not None:
                    raise ValueError('Duplicated • in: ' + str(self))
                self.pop(index)
                self.position = index


def read_rule(rule):
    """
    Return a list of symbols, i.e. (non-)terminals from a string
    (which has already been split to the left/right side)
    """
    if rule.strip() == 'ε':
        return

    attrs = None  # not None ⇔ inside the symbol
    head = False  # expecting the phrase head (or nothing)
    attr = ''  # a = plus an attribute value may follow
    equal_sign = False  # we are expecting a value for an attribute after =
    closed = False  # the symbol has already ended with )

    for token in shlex(rule, posix=True):
        if token == '(' or (token == '•' and (attrs is None or closed)):
            if attrs is not None:
                yield attrs
                if not closed:  # TODO: allow ambiguous (a=( b=1) ?
                    raise ValueError('Unexpected "("')

            if token == '•':
                yield '•'
                attrs = None
                continue

            attrs = Symbol()
            head = closed = False
        elif token == '?':
            attrs['optional'] = True
        elif token == ')':
            if equal_sign:
                raise ValueError('Unterminated attribute ' + attr)
            elif attr:
                attrs[attr] = None  # není lepší pro shodové atributy True?

            attr = ''
            equal_sign = False
            closed = True
        elif closed:
            raise ValueError('Not expected outside a symbol: ' + token)
        elif not attr:
            if token == '=':
                raise ValueError('Unexpected "="')
            elif not head and token[0].isupper():
                head = attrs['phrase'] = token
                continue
            attr = token
        elif token == '=':
            equal_sign = True
        elif equal_sign:
            attrs[attr] = token
            attr = ''
            equal_sign = False
        else:
            attrs[attr] = None
            attr = token

    if attrs is not None:
        yield attrs
        if not closed:
            raise ValueError('Rule not terminated by ")"')


need_escaping = re.compile(r'[()=? \t\n]').search  # podle shlex._find_unsafe


def escape(s):  # začátek podle shlex.quote
    if not isinstance(s, str):
        s = str(s)
    if not s:
        return "''"
    elif need_escaping(s) is None:
        return s
    elif "'" not in s:
        return "'" + s + "'"
    elif '"' not in s:
        return '"' + s + '"'
    else:
        raise ValueError("Can't escape, the string includes both quotes: " + s)
